import json
import logging
import os
from argparse import Namespace

import apiclient.discovery
import httplib2
from oauth2client import tools
from oauth2client.client import HttpAccessTokenRefreshError, flow_from_clientsecrets
from oauth2client.file import Storage
from oauth2client.service_account import ServiceAccountCredentials

from . import utils
from .exceptions import InvalidAuthException

logger = logging.getLogger(__name__)
logging.getLogger("googleapiclient").setLevel(logging.ERROR)
logging.getLogger("oauth2client").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.WARNING)

services = {}
keyfile_directory = []


def get_service_account_keyfile():
    keyfiles = []
    for loc in utils.KEYFILE_PATHS:
        for file in os.listdir(loc):
            if (
                file.endswith(".json") and file != "client_secret.json"
            ):  # TODO check if it has keyfile in it, not just any json
                return os.path.join(loc, file)
    raise InvalidAuthException(
        "No keyfile found. "
        "Make sure to pass your keyfile as a parameter or add it to ~/.agm"
    )


DEFAULT_USER = "unnamed"
CREDS_FOLDER = os.path.join(utils.AGM_PATH, "oauth_credentials")
if not os.path.exists(CREDS_FOLDER):
    os.makedirs(CREDS_FOLDER)


def get_client_id():
    return


def authenticate_client_secrets(scopes, user=None, noauth_local_webserver=False):
    if not user:
        user = DEFAULT_USER
    creds_path = os.path.join(CREDS_FOLDER, "{}.json".format(user))  # DRY
    store = Storage(creds_path)
    client_secret = os.path.join(utils.AGM_PATH, "client_secret.json")
    if not os.path.exists(client_secret):
        print(
            "Error: No client secret file found. Download a client secret and put it in {}".format(
                client_secret
            )
        )
        return
    flow = flow_from_clientsecrets(client_secret, scopes)
    # We have to pass fake args here because it wants to be run as a CLI
    fake_args = Namespace(
        auth_host_name="localhost",
        noauth_local_webserver=noauth_local_webserver,
        auth_host_port=[8080, 8090],
        logging_level="ERROR",
    )
    credentials = tools.run_flow(flow=flow, storage=store, flags=fake_args)
    logger.info("Authenticated client secrets and stored in {}".format(utils.AGM_PATH))
    return credentials


def get_oauth_credentials(user):
    if not user:
        return None
    creds_path = os.path.join(CREDS_FOLDER, "{}.json".format(user))
    if os.path.exists(creds_path):
        return Storage(creds_path).get()
    else:
        logger.debug(
            "No oauth credentials found for {} in {}".format(user, utils.AGM_PATH)
        )
        return None


# TODO: rename variables. add docs
class Service:
    def __init__(
        self,
        service_name,
        version,
        scopes_to_try=None,
        keyfile=None,
        delegated_email=None,
        all_methods=None,
    ):
        self.name = service_name
        self.version = version
        self.keyfile = keyfile
        self.scopes_to_try = scopes_to_try
        self.delegated_email = delegated_email

    @classmethod
    def from_individual_oauth_key(cls, *args, **kwargs):
        cls.__init__(*args, **kwargs)
        # TBD
        return

    @classmethod
    def from_service_account_key(cls, *args, **kwargs):
        cls.__init__(*args, **kwargs)
        # TBD
        return

    def get_service_account_service(self):
        for scope in self.scopes_to_try:
            service = self.attempt_authorization(self.keyfile, scope)
            if service:
                return service

    def attempt_authorization(self, keyfile, scope):
        for_user = " for user {}".format(self.delegated_email)
        try:
            credentials = ServiceAccountCredentials.from_json_keyfile_name(
                keyfile, scope
            )
            debug_line = "Authorized {} with scope {}".format(keyfile, scope)
            # delegated permissions
            if self.delegated_email:
                credentials = credentials.create_delegated(self.delegated_email)
                debug_line += for_user
            http = credentials.authorize(http=httplib2.Http())
            service = apiclient.discovery.build(self.name, self.version, http=http)
            logger.debug(debug_line)
            return service
        except HttpAccessTokenRefreshError:
            debug_line = "Failed to authorize keyfile {} using scope {}".format(
                keyfile, scope
            )
            if self.delegated_email:
                debug_line += for_user
            logger.debug(debug_line)
            return None

    def get_available_scopes(self):
        # slow
        # print
        return

    def get_oauth_service(self, credentials):  # DRY
        http = credentials.authorize(http=httplib2.Http())
        service_auth = apiclient.discovery.build(
            self.name, self.version, http=http
        )  # ??
        return service_auth

    def build(self):
        oauth = get_oauth_credentials(self.delegated_email)
        if oauth:
            logger.debug(
                "Got service for {} using individual oauth".format(self.delegated_email)
            )
            return self.get_oauth_service(oauth)
        else:
            self.keyfile = self.keyfile or get_service_account_keyfile()
            service = self.get_service_account_service()
            return service


def print_info():  # colorize this and format it nice
    output = {}
    output["oauth_creds"] = []
    creds_path = CREDS_FOLDER
    for credsfile in os.listdir(creds_path):
        with open(os.path.join(CREDS_FOLDER, credsfile)) as f:
            creds_info = json.load(f)
            user = credsfile.split(".json")[0]
            output["oauth_creds"].append(
                {"user": user, "scopes": creds_info.get("scopes")}
            )

    output["service_accounts"] = []
    try:
        keyfile = get_service_account_keyfile()
    except InvalidAuthException:
        keyfile = None
    if keyfile:
        with open(keyfile) as f:
            key_info = json.load(f)
            project = key_info.get("project_id")
            if project:
                output["service_accounts"].append(
                    {"project": project, "client_id": key_info["client_id"]}
                )
    print(json.dumps(output, indent=2))



#!/usr/bin/env python
"""
The main module -- parses APIs and builds commands to be sent to executor.py to
be ran. Could be refactored / broken up
"""
import argparse
import ast
import json
import logging
import logging.handlers
import os
import re
import sys
from distutils.version import StrictVersion
import requests

import pkg_resources

from tqdm import tqdm

from . import docs, executor, google_auth, utils
from .exceptions import InvalidCommandException, InvalidAuthException

logger = logging.getLogger(__name__)

# TODO USE CONFIGURATION if I can think of a compelling use case
# https://stackoverflow.com/questions/7567642/where-to-put-a-configuration-file-in-python


def get_parser():
    parser = argparse.ArgumentParser(
        description=utils.LOGO
        + "\nversion: "
        + pkg_resources.require("agm")[0].version,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument("-v", "--verbose", help="longer output", action="store_true")

    api = parser.add_argument_group(
        "API", "Call the Google API or get docs with agm [service] [resource] [method]"
    )
    api.add_argument(
        "command",
        help="The name of the service to call e.g. drive files list.",
        nargs="*",
    )

    api.add_argument(
        "-u", "--user", "--users", help="The impersonated user(s).", nargs="*"
    )
    api.add_argument("--outfile", help="Send output to a given file.")
    api.add_argument(
        "-d",
        "--docs",
        help="Open the documentation for this function.",
        action="store_true",
    )
    api.add_argument(
        "--version", help="specify api version, ie 'v2'. Optional", action="store_true"
    )
    api.add_argument(
        "--scopes",
        help="API scopes to give the key. Optional and usually unnecessary. "
        "If not specified, AGM will try all possible scopes that could authenticate "
        "a command until it finds one that works. You can drop everything before the "
        "last slash in the url, eg just use gmail.readonly",
        nargs="+",
    )
    api.add_argument(
        "--keyfile",
        help="Optional. specify the keyfile to authenticate with. By default, AGM"
        "looks for a .json file inside ~/.agm",
    )

    tools = parser.add_argument_group("Tools", "AGM tools")
    tools.add_argument("-V", help="AGM version", action="store_true")
    tools.add_argument(
        "--run_oauth",
        help="Authenticate an individual Google account and store the credentials."
        "Requires scopes and user to be set.",
        action="store_true",
    )
    tools.add_argument(
        "--noauth_local_webserver", help=argparse.SUPPRESS, action="store_true"
    )
    tools.add_argument(
        "--authinfo",
        help="Print information about authenticated keys",
        action="store_true",
    )
    return parser


def parse_known_args(args):
    parser = get_parser()
    return parser.parse_known_args(args)


def load_json_string_if_possible(input_string):
    """
    Parses input as a piped dictionary, if possible. Returns None otherwise.
    Tries standard json and jsonlines
    """
    output = None
    input_string = str(input_string)
    if input_string:
        try:
            output = json.loads(input_string)
            if isinstance(output, dict):
                output = [output]
            elif isinstance(output, list):
                pass
            else:
                output = None
        except Exception as e:  # TODO: specify exception
            lines = re.findall("[^}]+}", input_string)
            output = []
            for line in lines:
                try:
                    output.append(json.loads(line))
                except Exception as e:
                    return None
    return output or None


def parse_unknown_args(unknown):
    """
    Parse the "unknown" args -- ie, args that should be part of the api parameters.
    Everything is parsed as a list, and we will parse this further later.
    """
    api_parser = argparse.ArgumentParser()

    class EmptyIsTrue(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            if len(values) == 0:
                values = True
            setattr(namespace, self.dest, values)

    for arg in unknown:
        if arg.startswith(("-", "--")):
            api_parser.add_argument(arg, nargs="*", action=EmptyIsTrue)
    return api_parser.parse_args(unknown)


def process_body(api_options):
    if api_options.get("body"):
        api_options["body"] = [
            ast.literal_eval(item) for item in api_options["body"]
        ]  # TODO: Cleanup
    return api_options


def cli():  # MAIN
    argslist = sys.argv[1:]
    # TODO: Cleanup
    stdin = None
    if not sys.stdin.isatty():
        try:
            stdin = sys.stdin.read()
        except Exception as e:
            pass
    try:
        fully_parsed_args = parse_all_args(stdin, argslist)
        run(fully_parsed_args)
    except (InvalidCommandException, InvalidAuthException, argparse.ArgumentError) as e:
        if executor.GoogleAPIRequest.pbar:
            executor.GoogleAPIRequest.pbar.close()
        if logger.level == logging.DEBUG:
            logger.exception(e)
        print("ERROR: " + str(e))

    except Exception as e:
        print(
            "Unexpected error, please report to https://github.com/Cloudbakers/agm/issues"
        )
        raise e


def check_for_updates():
    local_version = pkg_resources.require("agm")[0].version
    url = "https://pypi.org/pypi/agm/json"
    remote_version = requests.get(url).json()["info"]["version"]
    if StrictVersion(remote_version) > StrictVersion(local_version):
        logger.warning(
            "WARNING: your build of AGM {} is out of date with the latest version {}. AGM is still in Alpha, so this may cause issues. Run pip install --upgrade agm to upgrade.".format(
                local_version, remote_version
            )
        )


def run(fully_parsed_args):
    args = fully_parsed_args[0]
    setup_logging(args.get("verbose"))
    logger.debug("Using options: {}".format(str(fully_parsed_args)))
    if args.get("V"):
        check_for_updates()
        print(pkg_resources.require("agm")[0].version)
        return
    elif args.get("authinfo"):
        google_auth.print_info()
        return
    elif args.get("run_oauth"):
        path = args.get("run_oauth")
        user = args.get("user")
        for item in ["scopes", "user"]:
            invalid = False
            if not args.get(item):
                print("Error! Please provide {}".format(item))
                invalid = True
        if invalid:
            return
        user = user[0]
        scopes = list(map(docs.parse_scopes, args.get("scopes")))
        google_auth.authenticate_client_secrets(
            scopes=scopes,
            user=user,
            noauth_local_webserver=args.get("noauth_local_webserver"),
        )
        return
    service, resource, method = docs.parse_command(
        args.get("command"), version=args.get("version")
    )
    if args.get("docs") or not method:
        docs.print_docs(service, resource, method)
        return
    scopes = (
        list(map(docs.parse_scopes, args.get("scopes") or [])) or method.get_scopes()
    )
    requests = get_list_of_requests(fully_parsed_args)
    executor.GoogleAPIRequest.pbar = tqdm(
        desc="progress",
        total=len(requests),
        unit="requests",
        smoothing=0.1,
        leave=False,
    )
    executor.multithread_execute_requests(
        requests, scopes=scopes, keyfile=args.get("keyfile")
    )
    executor.GoogleAPIRequest.pbar.close()


# TODO: Cleanup
def get_list_of_requests(fully_parsed_args):
    """
    """
    # two paths -- list or dict
    request_list = []
    for request_cmd in fully_parsed_args:
        command = request_cmd["command"]
        service_docs = docs.ServiceDocumentation(command[0])
        service = service_docs.name
        outfile = request_cmd.get("outfile")
        version = request_cmd.get("version") or service_docs.version
        reserved, _ = parse_known_args([])
        reserved = set(vars(reserved).keys()) - set(["user", "version"])
        nonreserved_params = {k: v for k, v in request_cmd.items() if k not in reserved}
        shared_params = {
            k: v[0]
            for k, v in nonreserved_params.items()
            if isinstance(v, list) and v is not None and len(v) == 1
        }
        tuple_list = [
            [(k, i) for i in v]
            for k, v in nonreserved_params.items()
            if isinstance(v, list) and v is not None and len(v) > 1
        ]
        if tuple_list:
            itr = zip(*tuple_list)
        else:
            itr = [()]
        for zipped in itr:
            parameters = {**dict(zipped), **shared_params}
            request_list.append(
                executor.GoogleAPIRequest(
                    service=service,
                    resources=command[1:-1],
                    method=command[-1],
                    version=version,
                    outfile=outfile,
                    **parameters,
                )
            )
    request_list.sort(key=lambda x: (x.user, x.version))
    return request_list


def parse_all_args(stdin, argslist):
    """
    """
    output = []
    all_options = {}
    if stdin:
        stdin = str(stdin)
    injson = load_json_string_if_possible(stdin)
    if stdin and not injson:
        argslist += stdin.splitlines()
    known, unknown = parse_known_args(argslist)
    all_options.update(vars(known))
    all_options.update(process_body(vars(parse_unknown_args(unknown))))
    if injson:
        outjson = []
        all_options_list = list(all_options.items())
        for item in injson:
            for k, v in item.items():
                if not isinstance(v, list):
                    item[k] = [v]
            # merge. Piped input overwrites command line flags if conflict
            outjson.append(dict(all_options_list + list(item.items())))
        return outjson
    else:
        output = [all_options]
    return output


class TqdmHandler(logging.StreamHandler):
    def __init__(self):
        logging.StreamHandler.__init__(self)

    def emit(self, record):
        msg = self.format(record)
        tqdm.write(msg, file=sys.stderr)


def setup_logging(verbose):
    """
    Set up logging with the TQDM handler and a rotating file
    handler in ~/.agm
    """
    logfolder = os.path.join(utils.AGM_PATH, "logs")
    if not os.path.exists(logfolder):
        os.makedirs(logfolder)
    logfile = os.path.join(logfolder, "agm_logs.log")
    file_handler = logging.handlers.RotatingFileHandler(
        logfile, mode="a", maxBytes=10 ** 7, backupCount=5
    )  # Keep ~50MB of logs. Could make this configurable
    file_handler.setLevel(logging.DEBUG)
    if verbose:
        level = logging.DEBUG
    else:
        level = logging.WARN
    logging.basicConfig(
        level=level,
        handlers=[TqdmHandler(), file_handler],
        format="%(asctime)s - %(name)s (%(lineno)s) - %(levelname)s: %(message)s",
        datefmt="%Y.%m.%d %H:%M:%S",
    )
    logger.debug("Setting logging with verbose = {}".format(verbose))


if __name__ == "__main__":
    cli()



import requests
import itertools
import os

LOGO = """
 _______ _______ __   __
|   _   |       |  |_|  |
|  |_|  |    ___|       |
|       |   | __|       |
|       |   ||  |       |
|   _   |   |_| | ||_|| |
|__| |__|_______|_|   |_|"""

# All the places to look for config files and json keys
AGM_PATH = os.path.expanduser("~/.agm")
KEYFILE_PATHS = [AGM_PATH]


def chunk_iter(iterable, n):
    it = iter(iterable)
    while True:
        chunk = tuple(itertools.islice(it, n))
        if not chunk:
            return
        yield chunk



"""
This is a module for interacting with the discovery document.
It's wonky and messy, definitely room for improvement
"""
import os
import textwrap
from datetime import timedelta

import colorful
import requests
import requests_cache

from . import utils, exceptions

colorful.use_style("solarized")
# Since the discovery document doesn't change often, I use a cache in order to
# Save time making calls to it.
# Warning! This will affect if you ever use the requests module elsewhere.

cache_path = os.path.join(utils.AGM_PATH, ".discovery_cache")
requests_cache.install_cache(cache_path, expire_after=timedelta(days=1))


def parse_scopes(name_or_url):
    """
    Allows for simpler specification of scopes instead of the full url
    """
    api_base_url = "https://www.googleapis.com/auth/"
    # This is the only exception I'm aware of
    gmail_url = "https://mail.google.com/"
    if name_or_url == "gmail" or name_or_url == gmail_url:
        name_or_url = "https://mail.google.com/"
    elif not name_or_url.startswith(api_base_url):
        name_or_url = api_base_url + name_or_url
    return name_or_url


def list_api_names_versions():
    result = requests.get(
        "https://www.googleapis.com/discovery/v1/apis",
        params={"fields": "items(name,version)"},
    )
    apis = result.json()["items"]
    return apis


def get_preferred_version(service_name):
    """ Get last nonbeta/nonalpha version, or earliest betas if its the only versions"""
    apis = [
        item for item in list_api_names_versions() if item.get("name") == service_name
    ]
    versions = [item.get("version") for item in apis]
    versions.sort(reverse=True)
    release_versions = [
        version
        for version in versions
        if ("alpha" not in version and "beta" not in version)
    ]
    if not release_versions:
        versions.reverse()
        if versions:
            return versions[0]
    else:
        return release_versions[0]


def print_apis():
    output_str = colorful.blue("Available APIs:")
    names = set()
    for api in list_api_names_versions():
        output_str += colorful.green("\n" + api.get("name") + " " + api.get("version"))
    print(output_str)


class ServiceDocumentation:
    def __init__(self, service, version=None):
        self.document = self.get_api_document(service, version)
        self.name = self.document["name"]
        self.version = self.document["version"]

    def get_all_scopes(self):
        return self.document["auth"]["oauth2"]["scopes"].keys()

    def get_api_document(self, api_name, version):
        # TODO: Retry logic
        # TODO: clean up a little
        # Admin APIs have a different naming scheme.
        if api_name in ["datatransfer", "directory", "reports"]:
            name = "admin"
            apis = list_api_names_versions()
            for api in apis:
                if api_name in api.get("version"):
                    version = api.get("version")
        else:
            name = api_name
            version = version or get_preferred_version(name)
        self.version = version
        result = requests.get(
            "https://www.googleapis.com/discovery/v1/apis/{}/{}/rest".format(
                name, version
            )
        )
        return result.json()

    def get_item_docs(self, item, indent=0):
        # Recursive
        result_str = ""
        if not item.get("properties"):
            return ""
        for property, value in item.get("properties").items():
            result_str += item_definition(
                property, value.get("type"), value.get("description"), indent
            )
            sub_item = None
            if value.get("items"):
                if value["items"].get("$ref"):
                    sub_schema = value["items"]["$ref"]
                    sub_item = self.document["schemas"][sub_schema]
            elif value.get("properties"):
                sub_item = value.get("items")
            if sub_item:
                result_str += self.get_item_docs(sub_item, indent=indent + 2)
        return result_str

    def get_resources(self):
        if self.document.get("resources"):
            return self.document["resources"].keys()

    def get_schema(self, schema_name):
        return self.document["schemas"][schema_name]

    def __str__(self):
        out_str = ""
        title = self.document.get("title")
        version = self.document.get("version")
        out_str += colorful.bold_red("{} {}\n".format(title, version))
        out_str += self.document.get("description") + "\n\n"
        out_str += header("Resources: ")
        for resource in self.get_resources():
            out_str += colorful.green(resource) + "\n"
        return out_str


class ResourceDocumentation:
    def __init__(self, service, resource_names):
        """
        Wonky
        """
        self.resource_names = resource_names
        self.service = service
        self.document = self.get_resource_docs(resource_names)
        pass

    def get_resource_docs(self, resource_names):
        """ eg calendar users """
        nested_resource = self.service.document
        name = None
        for resource in resource_names or []:
            child_resource = nested_resource.get("resources", {}).get(resource)
            if child_resource:
                name = resource
                nested_resource = child_resource
        self.name = name
        resource = nested_resource
        return resource

    def get_resources(self):
        if self.document.get("resources"):
            return self.document["resources"].keys()
        return []

    def get_methods_string(self, indent=0):
        resource = self.document  # rename
        methods = resource.get("methods")
        methods_string = ""
        if methods:
            methods_string += colorful.blue("Methods:")
            for k, v in methods.items():
                method = (
                    "\n"
                    + colorful.green(k)
                    + ": "
                    + wrapped_text(v.get("description"), indent)
                )
                methods_string += method
        return methods_string

    def __str__(self):
        out_str = ""
        out_str += header("Resource {}".format(".".join(self.resource_names)))
        out_str += "\n"
        if self.get_resources():
            out_str += header("Resources: ")
            out_str += colorful.green("\n".join(self.get_resources()))
            out_str += "\n\n"
        out_str += self.get_methods_string()
        return out_str


class MethodDocumentation:
    def __init__(self, service, resource, method_name, version=None):
        self.service = service
        self.service_docs = service.document
        self.resource_docs = resource.document
        self.docs = self.resource_docs["methods"].get(
            method_name
        )  # TODO: rename to document
        if not self.docs:
            raise exceptions.InvalidCommandException("Invalid method name")
        self.name = self.docs["id"]
        self.scopes = self.docs.get("scopes")

    def get_scopes(self):
        return self.docs.get("scopes")

    def get_request_body(self):
        request_body = self.docs.get("request")
        if request_body:
            schema = request_body.get("$ref")
            return self.service.get_schema(schema)

    def get_body_properties(self):
        body = self.get_request_body()
        if body:
            return body.get("properties")

    def __str__(self):
        out_str = header("Method: {}".format(self.name))
        out_str += self.docs.get("description")
        if self.docs.get("parameters"):
            out_str += "\n\n"
            out_str += header("Parameters:")
            for k, v in self.docs["parameters"].items():
                doc = item_definition(k, v["type"], v["description"])
                if v.get("required"):  # TODO: move these to be first
                    doc = colorful.red("(Required) ") + doc
                out_str += doc
        if self.get_request_body():
            out_str += header("\nRequest body:")
            out_str += self.service.get_item_docs(
                self.service.document, self.get_request_body()
            )
        if self.docs.get("response"):
            out_str += header("\nResult: ")
            out_str += self.service.get_item_docs(
                self.service.document["schemas"][self.docs.get("response")["$ref"]]
            )
        out_str += header("\nRequires one of these scopes:")
        out_str += "\n".join(self.docs.get("scopes"))
        return str(out_str)


def header(string):  # style a header
    return colorful.blue(string) + "\n"


def wrapped_text(string, indent=0):
    if not string:
        return ""
    lines = string.splitlines()
    result = ""
    result += textwrap.fill(lines[0], width=150, subsequent_indent=" " * (indent + 2))
    for line in lines[1:]:
        result += "\n"
        result += textwrap.fill(
            line,
            width=150,
            initial_indent=" " * (indent + 2),
            subsequent_indent=" " * (indent + 2),
        )
    return result


def item_definition(item, type, description, indent=0):
    a = " " * indent
    return a + "{}{}{} ({}{}{}): {}".format(
        colorful.green,
        item,
        colorful.reset,
        colorful.cyan,
        type,
        colorful.reset,
        wrapped_text(description, indent=indent) + "\n",
    )


def get_method_docs(
    service_name, resource_names, method_name, version=None
):  # awkward function
    service = ServiceDocumentation(service_name, version)
    resource = ResourceDocumentation(service, resource_names)
    method = MethodDocumentation(service, resource, method_name)
    return method


def parse_command(command, version=None):
    """
    Take a command and determine the service, resource and method
    """
    command_list = command
    service = None
    resource = None
    method = None
    try:
        if len(command_list) > 0:
            service_name = command_list[0]
            service = ServiceDocumentation(service_name, version)
        if len(command_list) > 1:
            resource = ResourceDocumentation(service, command_list[1:])
            if resource.name != command_list[-1]:
                method = MethodDocumentation(service, resource, command_list[-1])
        return service, resource, method
    except KeyError:
        raise exceptions.InvalidCommandException(
            "Invalid service, resource or method name."
        )


def print_docs(service, resource, method):
    # This is wonky
    if service and resource and method:
        print(method)
    elif service and resource:
        print(resource)
    elif service:
        print(service)
    else:
        print_apis()
